package com.sun.org.apache.bcel.internal.generic;

import com.sun.org.apache.bcel.internal.util.ByteSequence;
import java.io.IOException;
import java.io.DataOutputStream;

public abstract class BranchInstruction extends Instruction implements InstructionTargeter
{
    private int index;
    private InstructionHandle target;
    private int position;
    
    BranchInstruction() {
    }
    
    protected BranchInstruction(short opcode, InstructionHandle target) {
        super(opcode, (short)3);
        this.setTarget(target);
    }
    
    @Override
    public void dump(DataOutputStream out) throws IOException {
        out.writeByte((int)super.getOpcode());
        this.index = this.getTargetOffset();
        if (!Instruction.isValidShort(this.index)) {
            throw new ClassGenException("Branch target offset too large for short: " + this.index);
        }
        out.writeShort(this.index);
        return;
    }
    
    protected int getTargetOffset(InstructionHandle _target) {
        if (_target == null) {
            throw new ClassGenException("Target of " + super.toString(true) + " is invalid null handle");
        }
        int t = _target.getPosition();
        if (t < 0) {
            throw new ClassGenException("Invalid branch target position offset for " + super.toString(true) + ":" + t + ":" + _target);
        }
        return t - this.position;
    }
    
    protected int getTargetOffset() {
        return this.getTargetOffset(this.target);
    }
    
    protected int updatePosition(int offset, int max_offset) {
        this.position += offset;
        return 0;
    }
    
    @Override
    public String toString(boolean verbose) {
        String s = super.toString(verbose);
        String t = "null";
        if (verbose) {
            if (this.target != null) {
                if (this.target.getInstruction() == this) {
                    t = "<points to itself>";
                }
                else if (this.target.getInstruction() == null) {
                    t = "<null instruction!!!?>";
                }
                else {
                    t = "" + this.target.getPosition();
                }
            }
        }
        else if (this.target != null) {
            this.index = this.target.getPosition();
            t = "" + this.index;
        }
        return s + " -> " + t;
    }
    
    @Override
    protected void initFromFile(ByteSequence bytes, boolean wide) throws IOException {
        super.setLength(3);
        this.index = bytes.readShort();
    }
    
    public final int getIndex() {
        return this.index;
    }
    
    public InstructionHandle getTarget() {
        return this.target;
    }
    
    public void setTarget(InstructionHandle target) {
        notifyTarget(this.target, target, this);
        this.target = target;
    }
    
    static void notifyTarget(InstructionHandle old_ih, InstructionHandle new_ih, InstructionTargeter t) {
        if (old_ih != null) {
            old_ih.removeTargeter(t);
        }
        if (new_ih != null) {
            new_ih.addTargeter(t);
        }
    }
    
    @Override
    public void updateTarget(InstructionHandle old_ih, InstructionHandle new_ih) {
        if (this.target == old_ih) {
            this.setTarget(new_ih);
            return;
        }
        throw new ClassGenException("Not targeting " + old_ih + ", but " + this.target);
    }
    
    @Override
    public boolean containsTarget(InstructionHandle ih) {
        return this.target == ih;
    }
    
    @Override
    void setOpcode(short opcode) {
        if (this.target == null) {
            super.setOpcode(opcode);
        }
        else {
            InstructionHandle t = this.target;
            this.setTarget(null);
            super.setOpcode(opcode);
            this.setTarget(t);
        }
    }
    
    @Override
    void dispose() {
        this.setTarget(null);
        this.index = -1;
        this.position = -1;
    }
    
    protected int getPosition() {
        return this.position;
    }
    
    protected void setPosition(int position) {
        this.position = position;
    }
    
    protected void setIndex(int index) {
        this.index = index;
    }
}
